
package Log::Log4perl::AutoCategorize;
use strict;
our $VERSION = 0.008;

use Carp;
use IO::File;

use Log::Log4perl;
use Log::Dispatch::Screen;
use Log::Dispatch::File;
use Log::Log4perl::Appender;

use YAML;
use Data::Dumper;
$Data::Dumper::Indent=1;
$Data::Dumper::Sortkeys=1;
$Data::Dumper::Terse=1;

use Devel::Size qw(total_size);
$Devel::Size::warn = 0;

###################
my $MyPkg = __PACKAGE__;
my $Alias = $MyPkg; # was 'error inducing default', but that failed basic tests
my $opt;

use vars qw($AUTOLOAD);
my %seenCat;	# collect categories seen
my %unseenCat;	# filled at compile by optimizer, cleared at END
my %cat2data;	# category => [ logger-handle, level, munged-name, enablement ]
my %fn2cat;	# munged-name => category.  Needed for init-and-watch, not built til then

my $dumper;	# bound to Data::Dumper() or YAML::Dump by first call of AUTOLOAD
my $defConf;	# default logger config.

# silence redefine errs (no warnings wont do it, cuz theyre const);
#BEGIN { local $SIG{__WARN__} = sub { return if /redefined/; carp(@_) }};

BEGIN {
    # cuz Logger is used at import time (ie early) by many other
    # modules, it must load its own config at compile-time.
    # use Logger ($conffile) might be possible with import routine.

    # allow notify, do before init();
  Log::Log4perl::Logger::create_custom_level("NOTIFY", "WARN");
  Log::Log4perl::Logger::create_custom_level("NOTICE", "WARN");

    # initialize options to control debug output
    $opt = {
	# flags checked in optimization phase
	v => 0,		# generic verbosity
	f => 0,		# found op-chains (starting with pushmark)
	#F => 0,		# found op-chains op-dump
	m => 0,		# matched op-chains (found & end with method-named)
	M => 0,		# matched op-chains op-dump
	j => 0,		# junk op-chains (subset of found)
	J => 0,		# junk op-chains op-dump
	d => 0,		# $op->dump() while examining opcodes
	s => 0,		# print opstack while scanning for ending method_named op
	x => 0,		# extreme debug
	r => 0,		# function renaming (munging)
	z => 0,		# $op->dump when done optimizing
	w => 0,		# log wrong ops in chain that matched
	D => 0,		# break in optimizer if debugging

	# flags checked in AUTOLOAD
	i => 0,		# AUTOLOAD invoked
	A => 0,		# AUTOLOAD args
	a => 0,		# AUTOLOAD use of caller() 
	b => 0,		# anon-sub build
	c => 0,		# logging category creation
	n => 0,		# no optimize (dont stash the method)
	y => 0,		# use YAML::Dump (default is Data::Dumper, TBD)

	l => 1,		# add level to category, ie: pkg.sub.level.line
	e => 0,		# print END results to stdout
	z => 0,		# print size of stuff at END
    };

    $defConf = q{
	 log4perl.rootLogger = DEBUG, A1
	 log4perl.appender.A1 = Log::Dispatch::Screen
	 log4perl.appender.A1.layout = PatternLayout
	 log4perl.appender.A1.layout.ConversionPattern = %d %c: %m%n
	 # create COVERAGE log
	 log4perl.Log.Log4perl.AutoCategorize.END = INFO, COVERAGE
	 log4perl.appender.COVERAGE = Log::Dispatch::File
	 log4perl.appender.COVERAGE.filename = ./test-coverage.txt
         log4perl.appender.COVERAGE.mode = write
	 log4perl.appender.COVERAGE.layout = org.apache.log4j.PatternLayout
	 log4perl.appender.COVERAGE.layout.ConversionPattern = (%d{HH:mm:ss.SSS}) %c: %m%n
	 };

}


########

sub import {
    my $pkg = shift;
    my (%args) = @_;

    print "OK importing\n" if $opt->{v};

    # Log::Log4perl::AutoCategorize->
    set_debug($args{debug}) if $args{debug};

    if ($args{alias}) {
	local $, = ", ";
	my ($cpkg) = (caller(0))[0];
	if ($opt->{v}) {
	    print "aliasing $pkg as $args{alias}\n";
	    print "importer: ", (caller(0))[0..2];
	}
	no strict 'refs';
	*{$cpkg.'::'.$args{alias}} = *{$pkg};
	*{$cpkg.'::'.$args{alias}.'::AUTOLOAD'} = *{$pkg.'::AUTOLOAD'};
	$Alias = delete $args{alias};
    }

    if ($args{initfile}) {
	print "initialize with file\n" if $opt->{v};
        Log::Log4perl->init_and_watch($args{initfile}, 10);
	delete $args{initfile};
    }
    # someday, doing both might work.  right now, now its either or.
    elsif ($args{initstr}) {
	print "initialize with string\n" if $opt->{v};
        Log::Log4perl->init(\$args{initstr});
	delete $args{initstr};
    }
    else {
	print "initializing Logger with default\n" if $opt->{v};
        Log::Log4perl->init(\$defConf);
    }
    #$pkg->export_to_level(1, %args); 
}

sub set_debug {
    my ($arg) = pop @_;
    my $allowed = join '', sort keys %$opt;
    my $bad;
    foreach my $letter (split //, $arg) {
	if (defined $opt->{$letter}) {
	    $opt->{$letter} = 1;
	} else {
	    $bad .= $letter;
	}
    }
    die "illegal debug option(s): $bad - allowed: $allowed\n" if $bad;
}

sub get_loglevel {
    # returns the level-string, should really query base for complete set
    $DB::single =1;
    return $1 if $_[0] =~ m/^(?:log_)?(debug|info|warn|error|fatal|notice)/;
    return 0;
}

###################
sub AUTOLOAD {
    print "args: ", Dumper \@_ if $opt->{A};
    my $cpkg = $_[0];

    (my $meth = $AUTOLOAD) =~ s/.*:://;
    return if $meth eq 'DESTROY';
    print "\ncalled_as = $cpkg.$meth\n" if $opt->{i};

    # test if meth is a legitimate logging level
    my $level = get_loglevel($meth);
    unless ($level) {
	# delegate if possible (this is why u subclass)
	if (Log::Log4perl->can($meth)) {
	    print "delegating to Log::Log4perl->$meth()\n" if $opt->{v};
	    shift @_;
	    Log::Log4perl->$meth(@_);
	}
	elsif (Log::Log4perl::AutoCategorize->can($meth)) {
	    print "delegating to Log::Log4perl::AutoCategorize->$meth()\n"
		if $opt->{v};
	    shift @_;
	    Log::Log4perl::AutoCategorize->$meth(@_);
	}
	else { carp "$meth is not a legitimate log-level\n" }
	return;
    }
    # $meth = $level;
    print "meth: $meth\n" if $opt->{i};

    # use FQ name as category
    my ($pkg,$file,$ln0,$ln1,$sub0,$sub1);

    if (($pkg,$file,$ln1,$sub1) = caller(1)) {
	print "1: $pkg,$file,$ln1,$sub1\n" if $opt->{a};
    }
    ($pkg,$file,$ln0,$sub0) = caller(0);
    print "0: $pkg,$file,$ln0,$sub0\n" if $opt->{a};

    # construct category, avoid AUTOLOAD sub-name
    my $cat = $sub1 || 'main.main';
    $cat .= ".$level" if $opt->{l};
    $cat .= ".$ln0";
    $cat =~ s/::/./g;

    if ($cat2data{$cat}) {
	my $catinc = 'a';
	$catinc++ while $cat2data{$cat.$catinc};
	$cat .= $catinc;
    } 
    print "cat: $cat\n" if $opt->{c};
    delete $unseenCat{$meth};

    my $log = Log::Log4perl->get_logger($cat);
    my $predicate = "is_$level";
    
    # is it runnable ?
    my $runnable;
    eval { $runnable = $log->$predicate() };
    if ($@) {
	die("logger: cant $predicate on $cat: $@");
	return;
    }
    # record everything we might need in delegate
    $cat2data{$cat} = [ $log, $level, $meth, $runnable ];
    # $fn2cat{$meth} = $cat; # needed to expire stale routines.

    unless ($dumper) {
	$dumper = \&Data::Dumper::Dump;
	#$dumper = \&YAML::Dump if $opt->{y};
    }
    # make the right anonymous sub, depending on runability
    # avoid closure on $cat; make it string literal ??
    my $code;
    if (not $runnable) {
	print "building disabled sub: $cpkg.$meth\n" if $opt->{b};
	$code = sub { $seenCat{"$cat"}-- };
    }
    else {
	print "building enabled sub: $cpkg.$meth\n" if $opt->{b};
	$code = sub { logitDumper ("$cat", @_) };
	#$code = sub { logitYAML   ("$cat", @_) } if $opt->{y};
    }
    # stash it
    unless ($opt->{n}) {
	no strict 'refs';
	*{__PACKAGE__.'::'.$meth} = $code; 
	*{$Alias.'::'.$meth} = $code unless $opt->{n};
    }
    printf "code size for $meth: %d\n", total_size($code) if $opt->{z};
    goto &$code;
}

sub logitDumper {
    # log the message to base logger, using Data::Dumper to handle refs
    my ($cat, $cls, @args) = @_;
    my @scalars;

    $seenCat{"$cat"}++;
    my ($logger, $level) = @{$cat2data{$cat}};

    eval {
	# pull leading scalars from @args to @scalars
	push @scalars, shift @args while @args and not ref $args[0];

	# stringify @scalars, and Dump refs
	$logger->$level( !@scalars ? () : "@scalars, ",
			 !@args    ? () : Dumper((@args==1) ? @args : [@args]));
    };
    if ($@) { carp("logger dump problem on $cat: $@") }
    return "logged";
}

sub logitYAML {}

#######

sub dump {

}

sub DESTROY {
    print "# observed logging categories:\n", Dumper(\%seenCat);
}

sub myDump {
    print "# observed logging categories:\n", Dumper(\%seenCat);
}

sub sizeUsed {
    # returns the size of the entire stash
    total_size(\%Log::Log4perl::AutoCategorize::);
}


END {
    my %cat2munged;
    $cat2munged{$_} = $cat2data{$_}[2] foreach keys %cat2data;

    # for some reason, these dont all get munged, ignore for now.
    $Alias->info("Seen Log Events:", \%seenCat);
    $Alias->info("un-Seen Log Events:", \%unseenCat);
    $Alias->info("cat2data:", \%cat2munged);

    if ($opt->{e}) {
	print "Seen Log Events:"	=> Dumper \%seenCat;
	print "un-Seen methods:"	=> Dumper \%unseenCat;
	print "cat2info:"		=> Dumper \%cat2munged;
    }
    if ($opt->{y}) {
	print "Seen Log Events:"	=> Dump(%seenCat);
	print "un-Seen methods:"	=> Dump(%unseenCat);
	print "cat2data:"		=> Dump(%cat2munged);
    }
    if ($opt->{z}) {

	# print size info.  Devel::Size does an incomplete job on CVs,
	# so theres insufficient value to these numbers to base
	# decisions upon.

	my (%fnsizes, %hashsizes, %stashsizes, $total);

	foreach my $fn (values %cat2data) {
	    $total += $fnsizes{$fn->[2]}
		= total_size(\&{"Log::Log4perl::AutoCategorize::".$fn->[2]});
	}
	printf "total size of stashed subs: %d\n", sizeUsed();

	no strict 'refs';
	$total = 0;
	foreach my $hashname (qw(seenCat unseenCat cat2data)) {
	    $total += $hashsizes{$hashname} = total_size(\%{$hashname});
	}
	printf "total size of my hashs: %d\n", $total;

	if ($opt->{Z}) {
	    $total = 0;
	    foreach my $stashitem (values %Log::Log4perl::AutoCategorize::) {
		$total += $stashsizes{$stashitem} = total_size($stashitem);
	    }
	    printf "total size of stash items: %d\n", $total;
	}
	print "size breakdown: ", Dumper \%fnsizes, \%hashsizes, \%stashsizes;
    }
}

###################

#{ package optimizer; no warnings 'redefine' }
my $munged = '00000';

# sub method_munger
use optimizer 'extend-c' => sub {
    my $opp = shift;

    # look for op-chains which start with pushmark & const == __PACKAGE__
    # Scan until method_named is reached, while keeping track of inner
    # stack manipulations (iow monitor balance of push, pop on @opstack)
    
    my $n1 = $opp->name();
    $opp = $opp->next();
    return if ref $opp eq 'B::NULL';

    my $n2 = $opp->name();
    # by Policy, use Class method invocation only, hence const
    return unless $n1 eq "pushmark" and $n2 eq "const";

    # Class method allows code to expect the const opcodes value to be
    # Log::Log4perl::AutoCategorize or $Alias.  All others end the chain.

    my $class = '';
    eval { $class = $opp->sv->PV };
    return unless $class eq $Alias or $class =~/^$MyPkg/;

    $DB::single = 1 if $opt->{D};
    $opp->dump if $opt->{d};

    # OK: weve seen 2 required ops; pushmark, const='Logger'.  Now we
    # track stack activity, so that nested meth_named ops dont
    # prematurely end the scan which guards the munge.

    my (@opchain, @opstack, $name);

    while (@opstack or $opp->name ne 'method_named') {

	return if ref $opp eq 'B::NULL';
	$opp = $opp->next();
	return if ref $opp eq 'B::NULL';
	
	push @opchain, $opp;
	$name = $opp->name;
	
	if ($name eq 'pushmark') {
	    push @opstack, $opp;
	    printf "pushed: %s\n", opNames(\@opchain) if $opt->{s};
	}
	if ($name =~ /refgen|entersub/) {
	    printf "popping: %s\n", opNames(\@opchain) if $opt->{s};
	    pop @opstack;
	}
    }
    printf "found op-chain: $class => %s\n", opNames(\@opchain) if $opt->{f};

    # this should be proper end of chain
    my ($meth) = $opchain[-1];

    unless ($meth->name eq 'method_named') {
	# this is a sign of problems.
	printf "junk op-chain: $class => %s\n", opNames(\@opchain) if $opt->{j};
	dumpchain(\@opchain) if $opt->{J};
	return;
    }

    printf "matched op-chain: $class => %s\n", opNames(\@opchain) if $opt->{m};
    dumpchain(\@opchain) if $opt->{M};

    my $fnname = $meth->sv->PV;
    unless (get_loglevel($fnname)) {
	print "$fnname is ineligible for munging\n" if $opt->{v};
	return;
    }
    # now do the munge
    print "func: $fnname\n" if $opt->{r};
    $munged++;
    $meth->sv->PV("${fnname}_$munged");
    print "munged func name: ${fnname}_$munged\n" if $opt->{r};

    {
	# record munged fnname, and where its called (to aid test-coverage review)
	no warnings 'uninitialized';
	$unseenCat{"${fnname}_$munged"} = join(',', (caller(0))[0..2]);
    }
    $meth->dump if $opt->{z};
};

sub opNames {
    # given ref to opchain, prints names
    my ($opchain, $extra) = @_;
    return join ' ', map $_->name(), @$opchain unless $extra;
    return join ' ', map {$_->name(), $_->$extra()} @$opchain;
}

sub dumpchain {
    # annotates $op->dump with stuff, to stdout
    my ($opchain, @msg) = @_;
    foreach (@$opchain) {
	printf STDERR @msg;
	$_->dump;
    }
}

1;
__END__
################### ###################

=head1 NAME

Log::Log4perl::AutoCategorize - a Log4perl wrapper which uses your
package and subroutine names to create logging categories.

=head1 ABSTRACT

Log::Log4perl::AutoCategorize is a wrapper module for Log::Log4perl;
it provides a different interface for (what I perceive as) better
usability.  I assume hereafter that youve read and understood its
documentation (hereafter referred to as baseclass or base)

There are several more mature alternatives which you should check out
for comparison;

  #1. search for Stealth Loggers in base POD
  use Log::Log4perl qw(:easy);
  Log::Log4perl->easy_init($ERROR);

  #2. new functionality, developed at approx same time as AutoCategorize
  use Log::Log4perl::Filter;

=head1 SYNOPSIS

  use Log::Log4perl::AutoCategorize
    (
     alias => 'Logger', # shorthand class-name alias
     # easy init methods (std ones available too)
     # maybe later (when base has #includes), 2nd will override 1st
     initfile => $filename,
     initstr => q{
	 log4perl.rootLogger=DEBUG, A1
	 # log4perl.appender.A1=Log::Dispatch::Screen
	 log4perl.appender.A1 = Log::Dispatch::File
	 log4perl.appender.A1.filename = ./mylog
	 log4perl.appender.A1.mode = write
	 log4perl.appender.A1.layout = PatternLayout
	 log4perl.appender.A1.layout.ConversionPattern=%d %c %m%n
	 # create COVERAGE log
	 log4perl.logger.Logger.END = INFO, COVERAGE
	 log4perl.appender.COVERAGE = Log::Dispatch::File
	 log4perl.appender.COVERAGE.filename = ./test-coverage.txt
         log4perl.appender.COVERAGE.mode = write
	 log4perl.appender.COVERAGE.layout = org.apache.log4j.PatternLayout
	 log4perl.appender.COVERAGE.layout.ConversionPattern = (%d{HH:mm:ss.SSS}) %c: %m%n
	 },
     );

  foreach (1..500) {
    Logger->warn($_);
    foo();
    A->bar();
    A::bar();
  }

  sub foo {
    foreach (1..20) {
	Logger->warn($_);
    }
  }

  package A;

  sub bar {
    my @d;
    foreach (1..20) {
	push @d, $_;
	Logger->warn($_,\@d);
    }
  }


=head1 DESCRIPTION

Start with a CAVEAT; you must use static method style to get efficient
operation.  The explanation is is deferred until after the features
are described.

=head2 AutoCategorization

Normal use of Log4perl requires that you define a hierarchy of logging
categories, and code them into your classes.  Because the logging
category is orthogonal to severity, and because your categories are
essentially unbounded, you can get detailed control over the logging
done by your application.

But if youre lazy like me, you dont want the drudgery of coming up
with lots of categories, and making sure theyre consistent and
informative.  You note that your class hierarchy could serve as a
logging category.

Log::Log4perl provides easy_init(), which achieves some simplification
of categories, but you are still faced with creating them once, and
using them explicitly and consistently.  (Caveats apply; I read the
POD, but havent used the facility, due to what I percieved as its
weaknesses).

This package constructs the $logcat dynamically, using the package
name, subroutine name, and line number.  Simplifying somewhat, the
code does:

    1. uses caller() to construct a category string dynamically.
    2. fetches the $logger singleton {$logger = get_logger($logcat)}
    3. tests if $logcat was a loggable event {$logger->is_$logcat()}
    4. records the invocation.
    5. instantiates a handler anonymous subroutine for CALLER which
       delegates to $logger if 3 is true (discussed later)
    6. calls handler


=head2 Test Coverage Results of your Application

Step 4 is done because its informative; all statements which would
invoke the logger are cataloged at compile time, and are counted
during execution.

This can serve as a good 1st estimate of test coverage of your
application code.  This coverage is summarized at END time.

(15:24:24.772) Logger.END.info.106: $Seen Log Events: = [
  {
    'A.bar.debug.55' => '-400',
    'A.bar.warn.54' => 400,
    'Log.Log4perl.AutoCategorize.END.info.201' => 1,
    'main.foo.warn.44' => 200,
    'main.info.32' => 10,
    'main.warn.31' => 10
  }
(15:24:24.772) Logger.END.info.106: $un-Seen Log Events: = [
  {
    'debug_00011' => 'main,probe.pl,52,Log::Log4perl::AutoCategorize::__ANON__,1,0,,,0,',
    'info_00010' => 'main,probe.pl,51,Log::Log4perl::AutoCategorize::__ANON__,1,0,,,0,'
  }
];

Note that A.bar.debug.55 above has a negative count. This indicates
that the logging activity was suppressed by the configuration, but the
code was still reached 400 times.

The UnSeen log is provided to identify statements not executed during
test.  No categorization is possible, cuz thats done at 1st
invocation, but caller() is used to provide some usable info to find
the code, and devise a test for it.


=head2 Filtering with Log::Log4perl::AutoCategorize

Filtering with this package is based upon Log::Log4perls ability to
control logging based upon the category, combined with the creation of
hierarchical category names.  Using part or all of the hierarchy, you
can enable or disable logging on a class, a method in that class,
particular logging levels with that method, or line by line.

 # for example, debug Frob.pm, but only info on Frob::nicate()
 log4perl.category.Frob = DEBUG
 log4perl.category.Frob.nicate.debug = INFO
 # But - wanna see debug of line 118
 log4perl.category.Frob.nicate.debug.118 = DEBUG

This enables debug logging accross the Frob class, but then overrides
that class-wide setting within nicate(), suppressing debug logging (by
setting the threshold to INFO).  It further overrides that setting by
enabling the debug statement on line 118.

Its probably a bit confusing at 1st to have debug = INFO, but this is
intentional; it tells you the logging level thats coded (so you dont
have to go back to the source everytime), and therefore what is done
if the config-line is commented out.

It is strongly recommended that you use the test-coverage log to get a
complete and accurate catalog of all logging statements, this may be
provided by default in future versions.  To further ease this, I may
add an option to print the catalog with the 'log4perl.category' prefix
included.


=head1 Other Features

Automatic categorization, logging coverage, and fine-grained control
of logging are the primary value-added features of this wrapper.
Other secondary features are;

=head2 Configuration Loading

If use Log::Log4perl::AutoCategorize is given an import list which
includes initstr => q{} or initfile => $filename, the string or file
is used to initialize Log::Log4perl.  You can still use the std way
(TBT).

=head2 Classname Aliasing

Because it would be inconvenient to type the full classname for every
statement, ex: Log::Log4perl::AutoCategorize->debug(), the package has
the ability to alias itself into a classname of your choice, ex:

    use Log::Log4perl::AutoCategorize ( alias => myLogger );
    myLogger->debug("ok doing it");

CAVEAT: This package expects you to use static-class invocation, this
makes it possible for the optimizer to operate at compile-time based
upon the class, rather than waiting until runtime.  Since optimizer.pm
is designed to operate at compile-time, this is a necessary constraint.

Because aliasing is used in optimization phase, it must be imported
via use, it cannot be deferred as in require, import.

=head2 Easy Initialization

You can initialize the base logger by any of the following;

  # as use parameters, least typing
  a. use Log::Log4perl::Autocategorize ( initstr  => $string );
  b. use Log::Log4perl::Autocategorize ( initfile => $filename );

  # assuming youve already used the package
  c. Log::Log4perl::AutoCategorize->init( $filename );
  d. Log::Log4perl::AutoCategorize->init( \$string );

  # you can cheat, and use the base class name
  e. Log::Log4perl->init( $filename );
  f. Log::Log4perl->init( \$string );

  # assuming youve aliased
  g. Logger->init( $filename );
  h. Logger->init( \$string );

  i. # no explicit initialization at all (theres a default config)

The default config writes to stdout, and includes a test-coverage
setup.  This config is used at import() time, unless a or b is
provided by you.  If you do c-h, the base is re-initialized, which
seems to work.

I hope at some point to provide init_modify(), which will overlay new
configuration on existing, rather than a full reinitialization.  This
will be implemented using the base\'s notional configuration include
mechanism.

=head2 SubClassing

As sayeth L<perltoot/Planning for the Future: Better Constructors>,
you make you package suitable for deriving new ones.  test
04_subclass.t verifies this, though it wasnt the 1st thing I got
working.

=head1 Optimization Details

The use of caller() to look up the stack at runtime gives the ability
to report the method-name as well as the __PACKAGE__ and __LINE__
(which Log::Log4perl can do, via Log::Log4perl::PatternLayout
paramters: %C and %L ).

Using caller() is computationally expensive for a logging subsystem
which is supposed to be efficient.  By using optimizer.pm, Ive been
able to restore the runtime speed and retain the new feature-set.

=head2 optimizer

I use Simon Cozen''s optimizer.pm to find and replace all occurrences - 

    of    Logger->somefunc()
    with  Logger->"somefunc_$uniqueID"()

The function-munging produces a unique function name for every lexical
occurrence of a call to Logger, and and allows AUTOLOAD (which handles
all these unique, undefined subroutines) to create a custom handler for each.

=head2 lightweight customized subroutines

Given that all the customized anonymous subroutines are instantiated
as either sub{} or sub{ logit() }, they all invoke a common logit()
routine that actually does the work.  The asubs are trivial logit()
wrappers, with the category stringified and held in code, and passed
into logit() when called.

This doesnt shrink the number of symbol table entries, but it
minimizes the size of the CVs hanging there.

To save more space, a common do_nothing() function could be used
instead of the current invocation counting do-nothing, at the cost of
losing test-coverage info.

=head1 AUTOLOAD

At runtime, calls to Logger->warn_XYZ() are encountered, and are
dispatched to AUTOLOAD.  The following describes step 5 given
previously.

In step 5, a subroutine is instantiated.  The body of the subroutine
is determined by the outcome of step 3; the body either does nothing,
or does the actual logging by delegating to Log::Log4perl.  Because
all future invocations of this unique function must be due to THIS
CALLER, everything is customized into the CODE, including a closure on
the base log-object singleton.

An important caveat; all code must be written using class method
invocation style.  This is necessary cuz the optimization phase relys
on seeing the package name in the opcode stream.  Using a blessed
variable would not work, cuz the blessed-ness of the variable couldnt
be known at optimization time (unless Im wrong).

=head1 Benchmarking

Ive done one Benchmark.pm based test (tshort.pl) which doesnt seem
quite right.  The more straightforward test shows more convincing
results.

    $jimc@harpo logger]$ time t1.pl -n > /dev/null

    real	1m30.860s
    user	1m28.246s
    sys	0m2.002s

    $jimc@harpo logger]$ time t2.pl

    ... debug output snipped ...

    real	0m52.229s
    user	0m49.363s
    sys	0m1.625s

This last benchmark implements above test in pure Log::Log4perl code.
As you can see, this package is competetive with Log::Log4perl itself,
at least when analogous features are used (that use is a bit forced,
you wouldnt NEED to do so all the time).

Nonetheless, its conceivable to get better than native performance if
internal short-circuits can be taken.

Please run perftests/timeit.sh and send me email, particularly if you
get wildly different results.


=head1 TODO

Actually use it myself on a sizable project.  My project is not yet
using the Optimization feature of this package; even with the AUTOLOAD
and caller() overhead, things run well.  Application functionality
before optimization as they say..

=head2 fix raft of subroutine redefined warnings

Warnings are issued (under -w) during compile-phase, I cant control
them.  I suspect that solution lies in warnings::register, or perhaps
in more primitive solutions like { package B::Generate; no warnings
'redefine' }.  If you can patch these away, Id be eternally grateful.

=head2 devise good way to control logging details

AutoCategorize uses Data::Dumper to render the structure of complex
data arguments.  Ive found it would be nice to selectively suppress
the Dump, and print just the strings.

Ideally, Id have a way via the config-file, to control whether
data-ref details are expanded or not.  Heres what occurs to me, but as
it proposes changes to the config-file syntax, it should be considered
highly speculative:

    log4perl.category.TestUtils.testcase_applies = DEBUG, 0

The 2nd parameter would control whether Data::Dumper was invoked, in
this case they would not be.  This scheme has advantage of having a
wide API; hashrefs etc can (in principle - and subject to reality
check by Gurus) be given as parameters.

There are ways to avoid changing the arguments of config-items, but
they involve new parameter spaces, and other changes to Log::Log4perl

    log4perl.category.TestUtils.testcase_applies.__LINE_NUMBER__.detail = 0
    # wild-card the line-number ??
    log4perl.category.TestUtils.testcase_applies.*.detail = 0
    log4perl.category.TestUtils.*.*.detail = 1   # use Dumper
    log4perl.category.TestUtils.*.*.detail = -20 # limit levels of detail in Dumper
    log4perl.argserializer.TestUtils.*.*.detail = Data::Dumper
    log4perl.argserializer.TestUtils.*.*.detail = YAML
    log4perl.argserializer.TestUtils.*.*.detail = sub { YAML->dump }


=head2 allow use of alternates to Data::Dumper

YAML comes to mind, as does Data::Dumper::Sortkeys = sub {...}; I took
a quick swipe at using YAML::Dump (-dy debug option), but its not
yielding expected results.

Its also possible to get various object and context sensitive
serializations, esp if combined with extra params above.


=head2 devise method-name space

This wrapper knows that base logging methods are one of (debug, info,
notice, warn, error), and enforces this; the munger() AUTOLOAD w.  This should be stripped of suffixes so it
doesnt delegate thru (incorrectly) to Log4perl (where it probably
results in an error).  Currently, method name is stripped of 'log_?'
prefixes.


A whole panoply of method suffixes are possible; a careful use of this
method-name-space could be used to specify many subtle (or not so
subtle) variations in behavior (ex: debug_dump, warn_nodump,
error_iftrue, info_iffalse), some of which should be relegated to
configuration options rather than being specd in code.


=head2 short-circuit internal checks

Its likely that several checks internal to Log::Log4perl can be
eliminated once its determined (at 1st invocation) that a log event
should actually do something.


=head2 init and watch

Log4perl has ability to watch the logfile for changes, and change its
operation accordingly.  This package must hook into that notification,
and use it to undefine the appropriate symbol-table entry, thus
allowing the AUTOLOAD mechanism to re-vivify it according to the new
configuration.  Obviously it would be nice to do this only to the
parts of the config that are different.

This feature will use the %cat2data mapping built at 1st invocation
to allow config changes to be mapped back to the munged method-name
that resolved to that category in the 1st place.


=head2 context sensitive logging

NDC and MDC are tantalizing features, they appear to hold some promise
for conditional logging; ie once an application error is detected, a
full dump of the NDC or MDC is produced.  Alternatively, once an
all-ok is determined, a stack of NDC logs can be discarded, and a
summary message issued.


=head1 More TODO

    find and kill bugs..
    write test cases..
    optimize more
    integrate with Log::Log4perl
    bundle for CPAN..

    help/feedback is welcome.


=head1 AUTHOR

    Jim Cromie <jcromie@cpan.org>

=head1 COPYRIGHT AND LICENSE

Copyright 2003 Jim Cromie

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 

(I always wanted to write that ;-)


=cut

###################################

OLD CODE


    if (0) {
    #uzv     <0> pushmark s
    #uzw     <$> const(PV "Logger") sM/BARE
    #uzx     <$> const(PV "something") sM
    #uzy     <$> gvsv(*_) s
    #uzz     <$> method_named(PVIV "warn_00003") 

    my ($mark,$pkg,@const) = @opchain;
    
    unless ($mark->name eq 'pushmark') {
	if ($opt->{w}) {
	    print "name not pushmark\n";
	    $mark->dump;
	}
	return;
    }
    print "at pushmark\n"  if $opt->{v} or $opt->{s};
    if ($opt->{v}) {
	$pkg->dump();
	print $pkg->name, ' ', $pkg->sv->PV, "\n";
    }

    # look for something like: <$> const(PV "Logger") sM/BARE

    unless ($pkg->name eq 'const'
	    and ($pkg->sv->PV eq $Alias
		 or $pkg->sv->PV eq __PACKAGE__ ))
    {
	if ($opt->{w}) {
	    print "pkg const not alias\n";
	    $pkg->dump;
	}
	return;
    }
    unless ($meth->name eq 'method_named') {
	$meth->dump if $opt->{w};
	return;
    }
}
